/*
 *  A_DICE                              Copyright (C) Kimmo Kulovesi 1998
 *                                                    arkhan@softhome.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  The R250 engine first written by W. L. Maier, as described by
 *  S. Kirkpatrick and E. Stoll, Journal of Computational Physics,
 *  40, p. 517 (1981). Modified by me.
 *
 *  Please see 'license.txt'.
 *
 *****************************************************************************/

#include "a_dice.h"

/*****************************************************************************
 * The Dice Rolling Simulator, Copyright (C) Kimmo Kulovesi 1997, 1998       *

   Simulates the rolling of dice, using an advanced random number generation,
   allowing for more "random" results, than with the "1 + rand() % X" syntax.
 *                                                                           *
 *****************************************************************************/

/*****************************************************************************
 * The default die-rolling functions                                         *
 *****************************************************************************/

unsigned long RollDice(const unsigned int num_dice, const unsigned int num_sides)
{
    int i;
    unsigned long result = 0;

    for (i = num_dice; i; i--)
        result += (((rgen() / 211) & 0777777) % num_sides) + 1;

    return result;
};

long RollDice_B(const int num_dice, const int num_sides)
{
    int i;
    unsigned long result = 0;

    if (num_sides == 0 || num_dice <= 0)
        return 0;
    else if (num_sides > 0)
    {
        for (i = num_dice; i; i--)
            result += (((rgen() / 211) & 0777777) % num_sides) + 1;
    }
    else
    {
        for (i = num_dice; i; i--)
            result -= (((rgen() / 211) & 0777777) % -num_sides) + 1;
    }

    return result;
};

int RollDie_B(const int num_sides)
{
    if (!num_sides)
        return 0;
    else if (num_sides > 0)
        return (((rgen() / 211) & 0777777) % num_sides) + 1;
    else
        return -((((rgen() / 211) & 0777777) % -num_sides) + 1);
};

/*****************************************************************************
 * Faster versions of the above, at a slight cost in "randomness"            *
 *****************************************************************************/

unsigned long RDice(const unsigned int num_dice, const unsigned int num_sides)
{
    int i;
    unsigned long result = 0;

    for (i = num_dice; i; i--)
        result += 1 + rgen() % num_sides;

    return result;
};

long RDice_B(const int num_dice, const int num_sides)
{
    int i;
    unsigned long result = 0;

    if (num_sides == 0 || num_dice <= 0)
        return 0;
    else if (num_sides > 0)
    {
        for (i = num_dice; i; i--)
            result += 1 + rgen() % num_sides;
    }
    else
    {
        for (i = num_dice; i; i--)
            result -= 1 + rgen() % (-num_sides);
    }

    return result;
};

int RDie_B(const int num_sides)
{
    if (!num_sides)
        return 0;
    else if (num_sides > 0)
        return (1 + (rgen() % num_sides));
    else
        return -(1 + (rgen() % -num_sides));
};

/******************************************************************************
 * R250 Pseudorandom Number Engine                                            *
 *****************************************************************************/

int rgen_index = -1;

unsigned int rgen_buffer[250] =
{
    15301,57764,10921,56345,19316,43154,54727,49252,32360,49582,
    26124,25833,34404,11030,26232,13965,16051,63635,55860,5184,
    15931,39782,16845,11371,38624,10328,9139,1684,48668,59388,
    13297,1364,56028,15687,63279,27771,5277,44628,31973,46977,
    16327,23408,36065,52272,33610,61549,58364,3472,21367,56357,
    56345,54035,7712,55884,39774,10241,50164,47995,1718,46887,
    47892,6010,29575,54972,30458,21966,54449,10387,4492,644,
    57031,41607,61820,54588,40849,54052,59875,43128,50370,44691,
    286,12071,3574,61384,15592,45677,9711,23022,35256,45493,
    48913,146,9053,5881,36635,43280,53464,8529,34344,64955,
    38266,12730,101,16208,12607,58921,22036,8221,31337,11984,
    20290,26734,19552,48,31940,43448,34762,53344,60664,12809,
    57318,17436,44730,19375,30,17425,14117,5416,23853,55783,
    57995,32074,26526,2192,11447,11,53446,35152,64610,64883,
    26899,25357,7667,3577,39414,51161,4,58427,57342,58557,
    53233,1066,29237,36808,19370,17493,37568,3,61468,38876,
    17586,64937,21716,56472,58160,44955,55221,63880,1,32200,
    62066,22911,24090,10438,40783,36364,14999,2489,43284,9898,
    39612,9245,593,34857,41054,30162,65497,53340,27209,45417,
    37497,4612,58397,52910,56313,62716,22377,40310,15190,34471,
    64005,18090,11326,50839,62901,59284,5580,15231,9467,13161,
    58500,7259,317,50968,2962,23006,32280,6994,18751,5148,
    52739,49370,51892,18552,52264,54031,2804,17360,1919,19639,
    2323,9448,43821,11022,45500,31509,49180,35598,38883,19754,
    987,11521,55494,38056,20664,2629,50986,31009,54043,59743
};

void Init_Dice(unsigned long seed)
{
    int i, j;
    unsigned int msb, mask;

    rgen_index = -1;

        /* Arkhan: I combined the two loops into one, since I saw no reason
         *         why the numbers should be looped through twice... *shrug*
         */

    for (i = 0; i < 250; i++)     /* Fill the r250 buffer with 15-bit values */
    {
        seed = seed * 0x015a4e35L + 1;
        rgen_buffer[i] = ((seed>>16) & 0x7fff);

        seed = seed * 0x015a4e35L + 1;
        if (((seed>>16) & 0x7fff) > 16384)
            rgen_buffer[i] |= 0x8000;
    }

    msb = 0x8000;       /* To turn on the diagonal bit   */
    mask = 0xffff;      /* To turn off the leftmost bits */

    for (i = 0; i < 16; i++)
    {
        j = 11 * i + 3;             /* Select a word to operate on        */
        rgen_buffer[j] &= mask;     /* Turn off bits left of the diagonal */
        rgen_buffer[j] |= msb;      /* Turn on the diagonal bit           */
        mask >>= 1;
        msb >>= 1;
    }
};

unsigned int rgen()
{
    int i;

        /* Arkhan: I changed much of this function's structure, in order
         *         to make it work faster.
         */

    if (rgen_index >= 147)
    {
        if (rgen_index >= 249)
        {
            rgen_index = 0;
            i = 103;
        }
        else
            i = ++rgen_index - 147;      /* Wrap pointer around */
    }
    else
        i = ++rgen_index + 103;

    return (rgen_buffer[rgen_index] ^= rgen_buffer[i]);
};
