#include <stdio.h>
#include <string.h>
#include <math.h>
#include "astral.h"
#include "a_world.h"

/*****************************************************************************
*                                                                            *
* Some general functions used by the Sun/Moon code.                          *
*                                                                            *
*****************************************************************************/

#define RADDEG          ( 180.0 / PI )
#define DEGRAD          ( PI / 180.0 )

    /* The trigonometric functions in degrees */

#define sind(x)         sin((x) * DEGRAD)
#define cosd(x)         cos((x) * DEGRAD)
#define tand(x)         tan((x) * DEGRAD)

#define atand(x)        (RADDEG * atan((x)))
#define asind(x)        (RADDEG * asin((x)))
#define acosd(x)        (RADDEG * acos((x)))
#define atan2d(y, x)    (RADDEG * atan2((y),(x)))

long round(double i)
{
    if (i >= 0)
    {
        if (i - (floor(i)) < 0.5)
            return (long)floor(i);

        return (long)(floor(i) + 1);
    }
    else
    {
        if (i - (floor(i)) < 0.5)
            return (long)(floor(i) + 1);

        return (long)floor(i);
    }
};

double deg_360( double i )
{
      return (i - 360.0 * floor( i * (1.0 / 360.0)));
};

double deg_180( double i )
{
      return( i - 360.0 * floor( i * (1.0 / 360.0) + 0.5 ) );
};

/*****************************************************************************
*                                                                            *
* Original routines from 'SUNRISET.C', (C) Paul Schlyter, 1989, 1992.        *
* Released to the public domain by Paul Schlyter, December 1992.             *
*                                                                            *
*       Ruthlessly eviscerated and modified by me on July 1. 1998.           *
*                                                                            *
*****************************************************************************/

/**********************************************************************/
/*                                                                    */
/*       Northern latitude positive, Southern latitude negative       */
/*                                                                    */
/* Return value:  0 = sun rises/sets this day, times stored at        */
/*                    *trise and *tset.                               */
/*               +1 = sun above the specified "horizon" 24 hours.     */
/*                    *trise set to time when the sun is at south,    */
/*                    minus 12 hours while *tset is set to the south  */
/*                    time plus 12 hours.                             */
/*               -1 = sun is below the specified "horizon" 24 hours   */
/*                    *trise and *tset are both set to the time when  */
/*                    the sun is at south.                            */
/*                                                                    */
/**********************************************************************/
int sun_rise (signed char day, signed char month, signed long year,
                 float lat, double eccentricity, double obliquity,
                 double yl, double rt, double unk1, double unk2, double unk3,
                 double unk4, float *trise, float *tset)
{
    double  d,
            x, y, z,
            sr,
            sRA,
            sdec,
            tsouth;

        /* Values outside this range don't fit into a double */

    while(year < 1801)
        year += 200;
    while(year > 2099)
        year -= 100;

        /* Compute mean solar time */

    d = (yl * year - ( ( 7 * (year + ( (month + 9) / 12 ) ) ) / 4 ) +
        ( (275 * month) / 9 )
        + day - 730530L)
        + 0.5;

        /* Compute Sun's ecliptical coordinates */

    y = deg_360( unk3 + unk1 * d );
    z = eccentricity - 1.151E-9 * d;

      /* Compute true longitude and radius vector */

    x = y + z * RADDEG * sind(y) * ( 1.0 + z * cosd(y) );
    y = sqrt( 1.0 - z * z ) * sind(x);
    x = (cosd(x) - z);

    sr = sqrt( x * x + y * y );                            /* Solar distance */
    tsouth = (atan2d( y, x )) + (unk2 + unk4 * d);

    if ( tsouth >= 360.0 )
        tsouth -= 360.0;

        /* Compute ecliptic rectangular coordinates (z=0) */

    x = sr * cosd(tsouth);
    y = sr * sind(tsouth);

      /* Compute obliquity of ecliptic (inclination of Earth's axis) */

    z = obliquity - rt * d;

      /* Convert to equatorial rectangular coordinates - x is uchanged */

    y = y * cosd(z);
    z = (y * sind(z));

      /* Convert to spherical coordinates */

    sRA  = atan2d( y, x );
    sdec = atan2d( z, sqrt( x * x + y * y) );

        /* Compute time when Sun is at south - in hours UT */

    tsouth = 12.0
            - deg_180(
                        (
                            deg_360
                            (
                                (
                                (180.0 + unk3 + unk2)
                                + (unk1 + unk4)
                                * d
                                )
                            )
                            + 180.0
                        )
                        - sRA
                    )
            / 15.0;

        /* Compute the diurnal arc that the Sun traverses to reach */
        /* the specified altitude altit: */

    z = (-35/60) - (0.2666 / sr);

    sRA = ( sind(z) - sind(lat) * sind(sdec) ) /
          ( cosd(lat) * cosd(sdec) );

    *trise = (float)tsouth;
    *tset  = (float)tsouth;

    if ( sRA >= 1.0 )               /* Sun always below altit */
    {
        return -1;
    }
    else if ( sRA <= -1.0 )
    {
        *trise -= (float)12.0;
        *tset  += (float)12.0;

        return +1;
    }

    sr = acosd(sRA) / 15.0;      /* The diurnal arc, hours */

    *trise -= (float)sr;
    *tset  += (float)sr;

    return 0;
};

/******************************************************************************
 *
 *   The code to calculate the phase of Luna, as percentages of surface
 *   illuminated, at a given date.
 *
 *  Arkhan: I found the basis for this code on the 'Net, and as it was without
 *          the name of author or any copyright notices, I converted it to my
 *          dark purposes with a number of modifications. I haven't really
 *          the slightest idea on what causes this thing to work, except for
 *          the obscene amount of calculations, but since it does seem to be
 *          quite accurate, I'll use it... Only works for Luna, of course,
 *          as I haven't the faintest idea on how to convert this for Yael.
 *
 *          The routine for calculating the phases of Yael are at the end of
 *          this file; rather simple, yes, as the same phase always occurs
 *          in the same day.
 *****************************************************************************/

double luna_phase(double since)
{
    double N, l, Lbd, M, Mm, MmP;

    N = 360 * since / earth_year_length;
    N = deg_360(N);

    M = N + solar_ecc_long - solar_ecc_perigree;
    M = deg_360(M);
    M = sin((M * PI / 180));

    Lbd = N + (360 / PI * earth_eccentricity * M) + solar_ecc_long;
    l = 13.1763966 * since + lunar_mn_long;

    Lbd = deg_360(Lbd);
    l   = deg_360(l);

    Mm = l - (0.1114041 * since) - lunar_mn_perigree;
    Mm = deg_360(Mm);

    N = 1.2739 * sin(((2 * (l - Lbd) - Mm)  * PI / 180));

    MmP = sin((Mm + N - (0.1858 * M) - (0.37 * M)) * PI / 180);

    Mm = l + N + (6.2886 * MmP) - (0.1858 * M) + (0.214 * (MmP * 2));

    return (50 * (1 - cos((((Mm + (0.6583 * sin(((2 * (Mm - Lbd)) * PI / 180))))
            - Lbd)) * PI / 180)));
};

int earth_moon_phase(signed char day, signed char month, signed long year)
{
    double      since = earth_day_of_the_year(day, month, year),
                phase,
                next_p;
    signed long i;

    if (year > epoch_b + ((epoch - epoch_b) / 2))
        i = epoch;
    else
    {
        i = epoch_b;
        since -= epoch_b_dif;
    }

    if (i > year)
    {
        while (i > year)
        {
            since -= (365 + is_leap_year(i));

            if (i == 1)
                i = -1;
            else
                i--;
        }
    }
    else
    {
        while (i < year)
        {
            since += (365 + is_leap_year(i));
            i++;
        }
    }
    phase  = luna_phase(since);
    next_p = luna_phase(since + 1);
    since  = luna_phase(since - 1);

        /* If the next phase is larger, then the moon is now waxing */

    if (phase < next_p)
    {
            /* If the previous phase is smaller, then it must be New Moon */

        if (since > phase)
            return 0;

           /* First Quarter */

        if (phase < 50 && next_p > 50)
        {
            if ((next_p - 50) > (50 - phase))
                return 50;
            phase = round(phase);
            if (phase == 50)
                return 49;
            return (int)(phase);
        }
        else if (phase > 50 & since < 50)
        {
            if ((phase - 50) < (50 - since))
                return 50;
            phase = round(phase);
            if (phase == 50)
                return 51;
            return (int)(phase);
        }

        phase = round(phase);

            /* Full moon cannot be now, as the next phase is larger */

        if (phase == 100)
            return 99;
        else if (phase == 0)
            return 1;

        return (int)(phase);
    }

        /* Waning moon */

    else
    {
            /* Full moon */

        if (phase > since)
            return -100;

            /* Last quarter */

        if (phase > 50 && next_p < 50)
        {
            if ((50 - next_p) > (phase - 50))
                return -50;
            phase = round(phase);
            if (phase == 50)
                return -51;
            return -(int)(phase);
        }
        else if (phase < 50 & since > 50)
        {
            if ((50 - phase) < (since - 50))
                return -50;
            phase = round(phase);
            if (phase == 50)
                return -49;
            return -(int)(phase);
        }

        phase = round(phase);

            /* New moon cannot be now, since next phase is smaller */

        if (phase == 0)
            return -1;
        else if (phase == 100)
            return -99;

        return -(int)(phase);
    }
};

int harn_moon_phase(signed char day, signed char month, signed long year)
{
    switch(day)
    {
        case 7:     return 50;
        case 15:    return -100;
        case 23:    return -50;
        case 30:    return 0;
        default:
                    if (day < 15)
                        return (int)((float)(100.0 / 15.0) * day);
                    else
                        return (int)(-(100.0 - ((float)(100.0 / 15.0) * (day - 15))));
    }
};

int earth_age_of_moon(signed char day, signed char month, signed long year)
{
    int         i, j = 0;
    signed char tday = day,
                tmonth = month;
    signed long tyear = year;

    i = earth_moon_phase(day, month, year);

    while (i != 0)
    {
        j++;

        if (tday > 1)
            tday--;
        else if (tmonth > 1)
        {
            tmonth--;
            tday = earth_num_of_days(tmonth, tyear);
        }
        else
        {
            if (tyear != 1)
                tyear--;
            else
                tyear = -1;

            tmonth = num_of_months;
            tday = earth_num_of_days(tmonth, tyear);
        }

        i = earth_moon_phase(tday, tmonth, tyear);
    }

    return j;
};

int harn_age_of_moon(signed char day, signed char month, signed long year)
{
    if (day > 29)
        return 0;
    else
        return day;
};
